package com.ca.chorus.teiid.translator.async;

import java.sql.Connection;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Stack;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

import org.teiid.CommandContext;
import org.teiid.language.AndOr;
import org.teiid.language.ColumnReference;
import org.teiid.language.Command;
import org.teiid.language.Comparison;
import org.teiid.language.Condition;
import org.teiid.language.DerivedColumn;
import org.teiid.language.Expression;
import org.teiid.language.Literal;
import org.teiid.language.QueryExpression;
import org.teiid.language.Select;
import org.teiid.metadata.Column;
import org.teiid.metadata.RuntimeMetadata;
import org.teiid.translator.DataNotAvailableException;
import org.teiid.translator.ExecutionContext;
import org.teiid.translator.ResultSetExecution;
import org.teiid.translator.ReusableExecution;
import org.teiid.translator.TranslatorException;

public class TimerExecution implements ResultSetExecution, ReusableExecution<Connection> {

	private Select cmd;
	private final ExecutionContext executionContext;
	private final TimerangeExecutionFactory factory;
	private long start = 0;
	private long stop = 0;
	private long period = 0;
	private List<Long> timeRow = null;
	private AtomicBoolean awakeMode = new AtomicBoolean(true);
    private String[] needColumns  = null;
    private Map<String, Object> conditionMap = new HashMap<String, Object>();
    private Timer timer = null;

	public TimerExecution(QueryExpression expr,
			ExecutionContext executionContext, RuntimeMetadata metadata,
			Object connection, TimerangeExecutionFactory factory) {
		this.cmd = (Select)expr;
		this.executionContext = executionContext;
		this.factory = factory;
        this.needColumns = getNeededColumns(cmd.getDerivedColumns());
        Condition cond = this.cmd.getWhere();
        if (null == cond)
        {
        	System.out.println("WHERE condition is not specified. Input parameters are set to default (0).");
        } else {
            updateConditionMap(cond);
            
            if (this.conditionMap.get(factory.START) instanceof Long)
            {
                this.start = (Long) this.conditionMap.get(factory.START);
            }

            if (this.conditionMap.get(factory.STOP) instanceof Long)
            {
                this.stop = (Long) this.conditionMap.get(factory.STOP);
            }
            
            if (this.conditionMap.get(factory.PERIOD) instanceof Long)
            {
                this.period = (Long) this.conditionMap.get(factory.PERIOD);
            }
            System.out.println("########## init timer...");
    		initTimer();
        }

	}

	private void initTimer()
	{
		if (null == timer)
		{
	  		new Timer().schedule(new TimerTask(){
				@Override
				public void run() {
					System.out.println("############## Fire!...");						
					executionContext.dataAvailable();
					awakeMode.set(true);
				}					
			}, period, period);			
		}
	}

		
	@Override
	public void execute() throws TranslatorException {
		System.out.println("##### Execute...");
		timeRow = new ArrayList<Long>(3);
		timeRow.add(start);
		timeRow.add(start += period);
		timeRow.add(period);				
	}


	@Override
	public List<?> next() throws TranslatorException, DataNotAvailableException {		
		try {
			System.out.println("##### deliver row:" + (null == timeRow ? " null" : Arrays.toString(timeRow.toArray())));				
			return timeRow;				
		} finally {
			if (null == timeRow) // && awakeMode.get())
			{
				awakeMode.set(false);
				System.out.println("##### going to sleep");	
		  		throw DataNotAvailableException.NO_POLLING;
			}
			timeRow = null;
		}
	}
		
	
	@Override
	public void close() {
		System.out.println("############### close");
	}

	@Override
	public void cancel() throws TranslatorException {
		System.out.println("############### cancel");
	}
	
    private static String[] getNeededColumns(List<DerivedColumn> select)
    {
        String[] cols = new String[select.size()];
        Iterator<DerivedColumn> iter = select.iterator();
        int index = 0;
        while (iter.hasNext())
        {
            DerivedColumn symbol = iter.next();
            Expression expr = symbol.getExpression();
            if (expr instanceof ColumnReference)
            {
                Column element = ((ColumnReference) expr).getMetadataObject();
                cols[index] = element.getName();
            }
            index++;
        }
        return cols;
    }

    private void updateConditionMap(Comparison comparison,
    		Map<String, Object> condMap)
    {
    	String filter = comparison.getLeftExpression().toString();
    	filter = filter.substring(filter.indexOf('.') + 1).toLowerCase();
    	Object value = ((Literal) comparison.getRightExpression()).getValue();
    	if (condMap.containsKey(filter) && (value instanceof Long))
    	{
    		Long oldValue = (Long) condMap.get(filter);
    		if (filter.equalsIgnoreCase(factory.START))
    		{
    			if (oldValue < ((Long) value))
    			{
    				return;
    			}
    		}
    		if (filter.equalsIgnoreCase(factory.STOP))
    		{
    			if (oldValue > ((Long) value))
    			{
    				return;
    			}
    		}
    	}
    	condMap.put(filter, value);
    }

    private void updateConditionMap(Condition condition)
    {

    	if (condition instanceof Comparison)
    	{
    		updateConditionMap(((Comparison) condition), this.conditionMap);
    		return;
    	}
    	Stack<Condition> nodes = new Stack<Condition>();
    	nodes.push(condition);
    	Condition currentNode;
    	while (!nodes.isEmpty())
    	{
    		currentNode = nodes.pop();
    		Condition rightCondition = ((AndOr) currentNode).getRightCondition();
    		if (rightCondition instanceof AndOr)
    		{
    			nodes.push(rightCondition);
    		}
    		if (rightCondition instanceof Comparison)
    		{
    			updateConditionMap(((Comparison) rightCondition),
    					this.conditionMap);
    		}
    		Condition leftCondition = ((AndOr) currentNode).getLeftCondition();
    		if (leftCondition instanceof AndOr)
    		{
    			nodes.push(leftCondition);
    		}
    		if (leftCondition instanceof Comparison)
    		{
    			updateConditionMap(((Comparison) leftCondition),
    					this.conditionMap);
    		}
    	}

    }

	@Override
	public void reset(Command c, ExecutionContext executionContext,
			Connection connection) {
		System.out.println("############### reset");
		
	}

	@Override
	public void dispose() {
		System.out.println("############### dispose");
		if (null != timer)
		{
			timer.cancel();			
		}
		timer = null;
	}


}
